-- | A module for propositional formulas.
module Formula where

import Prelude hiding (and, not, or)

-- | /Propositional atoms/ are 'String's.
type Id = String

-- | A type of /propositional formulas/ consisting of bottom/false (ASCII input
-- notation @#@), propsitional atoms, and implication (ASCII input notation @->@).
data Formula = Bot | Atm Id | Imp Formula Formula
  deriving Eq

-- | Negation.
not :: Formula -> Formula
not f = f `Imp` Bot

and, or :: Formula -> Formula -> Formula
-- | Disjunction.
or f g = not f `Imp` g
-- | Conjunction.
and f g = not (f `Imp` not g)

-- | @(f -> g -> \#) -> \#@ is turned into @Just (f, g)@.
viewAnd, viewOr :: Formula -> Maybe (Formula, Formula)
viewAnd (Imp (Imp f (Imp g Bot)) Bot) = Just (f, g)
viewAnd _ = Nothing

-- | @((f -> #) -> g)@ is turned into @Just (f, g)@.
viewOr (Imp (Imp f Bot) g) = Just (f, g)
viewOr _ = Nothing

-- | Unicode output for formulas.
instance Show Formula where
  -- formulas
  showsPrec _ Bot = showString "⊥" -- bottom
  showsPrec _ (Atm x) = showString x -- atoms
  showsPrec d (Imp (Imp f (Imp g Bot)) Bot) = showParen (d > andPrec) $ -- conjunction
    showsPrec (andPrec + 1) f . showString " ∧ " . showsPrec andPrec g
    where andPrec = 6
  showsPrec d (Imp f Bot) = -- negation
    showString "¬ " . showParen (d > negPrec) (showsPrec negPrec f)
    where negPrec = 7
  showsPrec d (Imp (Imp f Bot) g) = showParen (d > orPrec) $ -- disjunction
    showsPrec (orPrec + 1) f . showString " ∨ " . showsPrec orPrec g
    where orPrec = 5
  showsPrec d (Imp f g) = showParen (d > impPrec) $ -- implication
    showsPrec (impPrec + 1) f . showString " → " . showsPrec impPrec g
    where impPrec = 4
  -- lists of formulas
  showList [] = id
  showList [f] = showsPrec 0 f
  showList (f:fs) = showsPrec 0 f . showString ", " . showList fs

-- | ASCII input syntax for propositional formulas.
-- 
--     * bottom/false: @#@
--     * negation: prefix @~@
--     * conjunction/and: right-associative infix @&@
--     * disjunction/or: right-associative infix @|@
--     * implication: right-associative infix @->@
instance Read Formula where
  readsPrec d r = [(Bot, s) | (x, s) <- lex r, x == "#"] -- bottom
    ++ readParen (d > negPrec) (\r -> -- negation
      [(not f, t) |
        ("~", s)  <- lex r,
        (f, t)    <- readsPrec negPrec s]) r
    ++ readParen (d > andPrec) (\r -> -- conjunction
      [(f `and` g, u) |
        (f, s)    <- readsPrec (andPrec + 1) r,
        ("&", t)  <- lex s,
        (g, u)    <- readsPrec andPrec t]) r
    ++ readParen (d > orPrec) (\r -> -- disjunction
      [(f `or` g, u) |
        (f, s)    <- readsPrec (orPrec + 1) r,
        ("|", t)  <- lex s,
        (g, u)    <- readsPrec orPrec t]) r
    ++ readParen (d > impPrec) (\r -> -- implication
      [(f `Imp` g, u) |
        (f, s)    <- readsPrec (impPrec + 1) r,
        ("->", t) <- lex s,
        (g, u)    <- readsPrec impPrec t]) r
    ++ [(Atm x, s) | (x, s) <- lex r, notReserved x] -- atoms
    where
      notReserved = (`notElem` ["","#","~","&","|","->"])
      negPrec = 7
      andPrec = 6
      orPrec  = 5
      impPrec = 4

